//---
//	gint:core:start - Kernel initialization and C runtime
//--

#include <gint/defs/attributes.h>
#include <gint/defs/types.h>
#include <gint/mmu.h>
#include <gint/drivers.h>
#include <gint/gint.h>
#include <gint/hardware.h>
#include <gint/exc.h>

#include "kernel.h"

/* Symbols provided by the linker script. For sections:
   - l* represents the load address (source address in ROM)
   - s* represents the size of the section
   - r* represents the relocation address (destination address in RAM)
   gint's BSS section is not mentioned here because it's never initialized */
extern uint32_t
	brom, srom,			/* Limits of ROM mappings */
	ldata,  sdata,  rdata,		/* User's data section */
	lilram, silram, rilram,		/* IL memory section */
	lxram,  sxram,  rxram,		/* X memory section */
	lyram,  syram,  ryram,		/* Y memory section */
	sbss, rbss;			/* User's BSS section */
#ifdef FX9860G
extern uint32_t
	lgmapped, sgmapped, rgmapped;	/* Permanently mapped functions */
#endif

/* Constructor and destructor arrays */
extern void (*bctors)(void), (*ectors)(void);
extern void (*bdtors)(void), (*edtors)(void);

/* User-provided main() function */
int main(int isappli, int optnum);

/* regcpy(): Copy a memory region using symbol information
   @l  Source pointer (load address)
   @s  Size of area (should be a multiple of 16)
   @r  Destination pointer (relocation address) */
static void regcpy(uint32_t * restrict l, int32_t s, uint32_t * restrict r)
{
	while(s > 0)
	{
		*r++ = *l++;
		*r++ = *l++;
		*r++ = *l++;
		*r++ = *l++;
		s -= 16;
	}
}
#define regcpy(l, s, r) regcpy(&l, (int32_t)&s, &r)

/* regclr(): Clear a memory region using symbol information
   @r  Source pointer (base address)
   @s  Size of area (should be a multiple of 16) */
static void regclr(uint32_t *r, int32_t s)
{
	while(s > 0)
	{
		*r++ = 0;
		*r++ = 0;
		*r++ = 0;
		*r++ = 0;
		s -= 16;
	}
}
#define regclr(r, s) regclr(&r, (int32_t)&s)

/* callarray(): Call an array of functions (constructors or destructors)
   @f  First element of array
   @l  First element outside of the array */
static void callarray(void (**f)(void), void (**l)(void))
{
	while(f < l) (*(*f++))();
}

/* start() - this is where it all starts
   Returns a status code. Invoking main menu is better than returning! */
GSECTION(".text.entry")
int start(int isappli, int optnum)
{
	/* We are currently in a dynamic userspace mapping of an add-in run
	   from the storage memory. We are running in privileged mode with one
	   golden rule:

	       Do not disturb the operating system.

	   gint will silently run in an isolated part of the memory (fx9860g)
	   or at the start of the RAM (fxcg50). The kernel will redirect
	   interrupts and load its own drivers, so we can't rely too much on
	   the system. Ladies and gentlemen, let's have fun! ;D */

	/* For now, we use the system's memory mapper for ROM. We'll still do
	   it later in our TLB miss handler once we're installed. RAM is always
	   fully mapped, but we need to initialize it. We also need to do some
	   hardware detection because old fx9860g models have a different
	   processor with some incompatible features. */

	/* Detect architecture; this will tell SH3 from SH4 on fx9860g */
	hw_detect();

	/* Load data sections and wipe the bss section. This has to be done
	   first for static and global variables to be initialized */
	regcpy(ldata,  sdata,  rdata);
	#ifdef FX9860G
	regcpy(lgmapped, sgmapped, rgmapped);
	#endif
	regcpy(lilram, silram, rilram);
	regcpy(lxram,  sxram,  rxram);
	regcpy(lyram,  syram,  ryram);
	regclr(rbss, sbss);

	/* Install gint, switch VBR and initialize drivers */
	kinit();

	/* We are now running on our own in kernel mode. Since we have taken
	   control of interrupts, pretty much any interaction with the system
	   will break it. We'll limit our use of syscalls and do device driving
	   ourselves. (Hopefully we can add cool features in the process!) */

	/* Now that we have initialized the kernel, we are ready to start the
	   hosted user application, which has its own constructors and
	   destructors to work with. */

	callarray(&bctors, &ectors);
	int ret = main(isappli, optnum);
	callarray(&bdtors, &edtors);

	/* Before leaving the application, we need to clean everything we
	   changed to hardware settings and peripheral modules. The OS is bound
	   to be confused (and hang, or crash, or any other kind of giving up)
	   if we don't restore them. */

	/* Unload gint and give back control to the system. Driver settings
	   will be restored while interrupts are disabled */
	kquit();

	/* TODO: Invoke main menu so that add-in can restart? */
	return ret;
}
